// $Date:: 2010-02-21 20:19:09 +0900#$
// $Rev: 2 $

#include "DotRemove.h"
#include <math.h>
#include <emmintrin.h>
#include <smmintrin.h>

//#define PRINTTIME
#ifdef PRINTTIME
#include <stdio.h>
#include <mmsystem.h>
#pragma comment ( lib , "winmm.lib" )
static __int32 tcount = 0 ;
#endif

/**************************************************************************************************/
/* DotRemove                                                                                      */
/**************************************************************************************************/
//==============================================================================
// constructor
//==============================================================================
DotRemove::DotRemove(PClip _child, int strength, float threshold, int type, int debug, IScriptEnvironment* env) :
    GenericVideoFilter(_child) ,
    Height      ( this->vi.height         ) ,
    Width       ( this->vi.width          ) ,
    Strength    ( strength                ) ,
    Threshold   ( (int) ( threshold * 4 ) ) ,
    Type        ( type                    ) ,
    Debug       ( debug                   )
{
#define C_CHACHE_LEN   ( ( this->Height + 2 ) * ( ( this->Width >> 3 ) + 2 ) * sizeof (__m128i) )
#define M_ALLOC(ptr_m) \
{ \
    ptr_m = ( (__int16*) _aligned_malloc ( C_CHACHE_LEN , 16 ) ) ; \
    ZeroMemory ( ptr_m , C_CHACHE_LEN ) ; \
    ptr_m += sizeof (__m128i) / 2 ; \
}
    M_ALLOC ( this->Y ) ;
    M_ALLOC ( this->avg ) ;
    M_ALLOC ( this->vag ) ;
    M_ALLOC ( this->hag ) ;
#undef M_ALLOC
    {
        unsigned int output_ecx;
        unsigned int output_edx;
        __asm{
            push eax
            push ebx
            push ecx
            push edx

            mov eax, 1

            CPUID //CPUID߂s(FPentiumȍ~K{)

            mov output_ecx, edx
            mov output_edx, edx

            pop edx
            pop ecx
            pop ebx
            pop eax
        }

        if ( ! ( output_edx & 0x04000000 ) || ( output_ecx & 0x00080001 ) != 0x00080001 )
            env->ThrowError("DOTREMOVE:this machine do not support SSE2/3/4.");
    }

    if ( strength < 0 || 128 < strength )
            env->ThrowError("DOTREMOVE:strength must be 0 to 128.");

    if ( threshold < 0 || 128 < threshold )
            env->ThrowError("DOTREMOVE:threshold must be 0 to 128.");

    if ( type < 0 || 2 < type )
            env->ThrowError("DOTREMOVE:type must be 0 to 2.");
}

//==============================================================================
// destructor
//==============================================================================
DotRemove::~DotRemove()
{
#define M_FREE(ptr_m) \
{ \
    ptr_m -= sizeof (__m128i) / 2 ; \
    _aligned_free ( ptr_m ) ; \
}
    M_FREE ( this->Y ) ;
    M_FREE ( this->avg ) ;
    M_FREE ( this->vag ) ;
    M_FREE ( this->hag ) ;
#undef M_FREE

#ifdef PRINTTIME
    char strtime [ 256 ] ;
    _snprintf ( strtime , sizeof (strtime) , "%f\r\n" , ( (float) tcount ) / 1000 ) ;
    OutputDebugString ( strtime ) ;
#endif // PRINTTIME
}

//==============================================================================
// GetFrame
//==============================================================================
PVideoFrame DotRemove::GetFrame (int n, IScriptEnvironment* env)
{
    PVideoFrame src = this->child->GetFrame ( n , env ) ;
    env->MakeWritable ( & src ) ;
    BYTE* pDst = src->GetWritePtr () ;
    const int nSrcPitch = src->GetPitch();
    static const __m128i Low16Mask = _mm_set1_epi16 ( 0x00FF ) ;
    const int OneLineSpan = ( this->Width + 16 ) >> 3 ;

#ifdef PRINTTIME
    tcount -= timeGetTime () ;
#endif // PRINTTIME

    {
        // l̎擾
        __m128i * d = (__m128i *) this->Y  + OneLineSpan ;
        for ( int i = 0 ; i < this->Height ; i ++ , d += 2 )
        {
            const __m128i * s = (const __m128i *) ( pDst + i * nSrcPitch ) ;

            // [
            _mm_store_si128 ( d - 1 , _mm_set1_epi16 ( ( (BYTE*) s ) [ 0 ] ) ) ;

            // 
            for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , s ++ )
                _mm_store_si128 ( d , _mm_and_si128 ( _mm_lddqu_si128 ( s ) , Low16Mask ) ) ;

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
        }
        // [
        CopyMemory ( d - 1 , d - OneLineSpan - 1 , OneLineSpan ) ;

        // [
        d = (__m128i *) this->Y  ;
        CopyMemory ( d - 1 , d + OneLineSpan - 1 , OneLineSpan ) ;
    }

    // PxώZo
    switch ( this->Type )
    {
    case 0:
        {
            // KEXڂ
            const __m128i * s0 = (const __m128i *) ( this->Y - 1 ) + OneLineSpan ;
            const __m128i * s1 = (const __m128i *) ( this->Y     ) + OneLineSpan ;
            const __m128i * s2 = (const __m128i *) ( this->Y + 1 ) + OneLineSpan ;
            __m128i * d = (__m128i *) this->vag + OneLineSpan ;
            for ( int i = 0 ; i < this->Height ; i ++ , d +=2 , s0 += 2 , s1 += 2 , s2 += 2 )
            {
                // [
                _mm_store_si128 ( d - 1 , _mm_add_epi16 ( _mm_lddqu_si128 ( s0 ) , _mm_add_epi16 ( _mm_slli_epi16 ( _mm_load_si128 ( s1 ) , 1 ) , _mm_lddqu_si128 ( s2 ) ) ) ) ;

                // 
                for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , s0 ++ , s1 ++ , s2 ++ )
                    _mm_store_si128 ( d , _mm_add_epi16 ( _mm_lddqu_si128 ( s0 ) , _mm_add_epi16 ( _mm_slli_epi16 ( _mm_load_si128 ( s1 ) , 1 ) , _mm_lddqu_si128 ( s2 ) ) ) ) ;

                // E[
                _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
            }
            // [
            CopyMemory ( d - 1 , d - OneLineSpan - 1 , OneLineSpan ) ;

            // [
            d = (__m128i *) this->vag ;
            CopyMemory ( d - 1 , d + OneLineSpan - 1 , OneLineSpan ) ;

            d = (__m128i *) this->avg + OneLineSpan ;
            s0 = (const __m128i *) this->vag + 0 * OneLineSpan ;
            s1 = (const __m128i *) this->vag + 1 * OneLineSpan ;
            s2 = (const __m128i *) this->vag + 2 * OneLineSpan ;
            for ( int i = 0 ; i < this->Height ; i ++ , d +=2 , s0 += 2 , s1 += 2 , s2 += 2 )
            {
                // [
                _mm_store_si128 ( d - 1 , _mm_add_epi16 ( _mm_load_si128 ( s0 ) , _mm_add_epi16 ( _mm_slli_epi16 ( _mm_load_si128 ( s1 ) , 1 ) , _mm_load_si128 ( s2 ) ) ) ) ;

                // 
                for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , s0 ++ , s1 ++ , s2 ++ )
                    _mm_store_si128 ( d , _mm_add_epi16 ( _mm_load_si128 ( s0 ) , _mm_add_epi16 ( _mm_slli_epi16 ( _mm_load_si128 ( s1 ) , 1 ) , _mm_load_si128 ( s2 ) ) ) ) ;

                // E[
                _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
            }
            // [
            CopyMemory ( d - 1 , d - OneLineSpan - 1 , OneLineSpan ) ;

            // [
            d = (__m128i *) this->avg ;
            CopyMemory ( d - 1 , d + OneLineSpan - 1 , OneLineSpan ) ;

            d = (__m128i *) this->avg - 1 ;
            for ( int i = 0 ; i < ( this->Height + 2 ) * ( ( this->Width >> 3 ) + 2 ) ; i ++ , d ++ )
            {
                _mm_store_si128 ( d , _mm_srli_epi16 ( _mm_load_si128 ( d ) , 4 ) ) ;
            }
        }
        break;

    case 1:
        {
            // 3x3 P
            const __m128i * s0 = (const __m128i *) ( this->Y - 1 ) + OneLineSpan ;
            const __m128i * s1 = (const __m128i *) ( this->Y     ) + OneLineSpan ;
            const __m128i * s2 = (const __m128i *) ( this->Y + 1 ) + OneLineSpan ;
            __m128i * d = (__m128i *) this->vag + OneLineSpan ;
            for ( int i = 0 ; i < this->Height ; i ++ , d +=2 , s0 += 2 , s1 += 2 , s2 += 2 )
            {
                // [
                _mm_store_si128 ( d - 1 , _mm_add_epi16 ( _mm_lddqu_si128 ( s0 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 ) , _mm_lddqu_si128 ( s2 ) ) ) ) ;

                // 
                for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , s0 ++ , s1 ++ , s2 ++ )
                    _mm_store_si128 ( d , _mm_add_epi16 ( _mm_lddqu_si128 ( s0 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 ) , _mm_lddqu_si128 ( s2 ) ) ) ) ;

                // E[
                _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
            }
            // [
            CopyMemory ( d - 1 , d - OneLineSpan - 1 , OneLineSpan ) ;

            // [
            d = (__m128i *) this->vag ;
            CopyMemory ( d - 1 , d + OneLineSpan - 1 , OneLineSpan ) ;

            d = (__m128i *) this->avg + OneLineSpan ;
            s0 = (const __m128i *) this->vag + 0 * OneLineSpan ;
            s1 = (const __m128i *) this->vag + 1 * OneLineSpan ;
            s2 = (const __m128i *) this->vag + 2 * OneLineSpan ;
            for ( int i = 0 ; i < this->Height ; i ++ , d +=2 , s0 += 2 , s1 += 2 , s2 += 2 )
            {
                // [
                _mm_store_si128 ( d - 1 , _mm_add_epi16 ( _mm_load_si128 ( s0 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 ) , _mm_load_si128 ( s2 ) ) ) ) ;

                // 
                for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , s0 ++ , s1 ++ , s2 ++ )
                    _mm_store_si128 ( d , _mm_add_epi16 ( _mm_load_si128 ( s0 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 ) , _mm_load_si128 ( s2 ) ) ) ) ;

                // E[
                _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
            }
            // [
            CopyMemory ( d - 1 , d - OneLineSpan - 1 , OneLineSpan ) ;

            // [
            d = (__m128i *) this->avg ;
            CopyMemory ( d - 1 , d + OneLineSpan - 1 , OneLineSpan ) ;

            d = (__m128i *) this->avg - 1 ;
            for ( int i = 0 ; i < ( this->Height + 2 ) * ( ( this->Width >> 3 ) + 2 ) ; i ++ , d ++ )
            {
                const __m128i wk = _mm_load_si128 ( d ) ;
                _mm_store_si128 (
                    d ,
                    _mm_srai_epi16 (
                        _mm_adds_epi16 ( _mm_adds_epi16 ( _mm_adds_epi16 (
                            _mm_mullo_epi16 ( _mm_adds_epi16 ( wk , _mm_sign_epi16 ( _mm_set1_epi16 ( 5 ) , wk ) ) , _mm_set1_epi16 ( 7 ) ) ,
                            _mm_srai_epi16 ( wk , 4 ) ) ,
                            _mm_srai_epi16 ( wk , 5 ) ) ,
                            _mm_srai_epi16 ( wk , 6 ) ) ,
                        6 ) ) ;
            }
        }
        break ;

    default:
        {
            // WAڂ
            const __m128i * s0 = (const __m128i *) ( this->Y     ) + OneLineSpan ;
            const __m128i * s1 = (const __m128i *) ( this->Y - 1 ) + OneLineSpan ;
            const __m128i * s2 = (const __m128i *) ( this->Y + 1 ) + OneLineSpan ;
            const __m128i * s3 = (const __m128i *) ( this->Y     ) + OneLineSpan * 0 ;
            const __m128i * s4 = (const __m128i *) ( this->Y     ) + OneLineSpan * 2 ;
            __m128i * d = (__m128i *) this->avg + OneLineSpan ;
            for ( int i = 0 ; i < this->Height ; i ++ , d +=2 , s0 += 2 , s1 += 2 , s2 += 2 , s3 += 2 , s4 += 2 )
            {
                // [
                const __m128i ss [ 5 ] = {
                        _mm_load_si128 ( s0 ) ,
                        _mm_lddqu_si128 ( s1 ) , 
                        _mm_lddqu_si128 ( s2 ) ,
                        _mm_load_si128 ( s3 ) ,
                        _mm_load_si128 ( s4 ) } ;
                const __m128i wk0 = _mm_subs_epi16 ( _mm_subs_epi16 (
                    _mm_adds_epi16 ( _mm_adds_epi16 ( _mm_adds_epi16 ( _mm_adds_epi16 ( 
                        ss [ 0 ] ,ss [ 1 ] ) , ss [ 2 ] ) , ss [ 3 ] ) , ss [ 4 ] ) ,
                    _mm_max_epi16 ( _mm_max_epi16 ( _mm_max_epi16 ( _mm_max_epi16 ( 
                        ss [ 0 ] ,ss [ 1 ] ) , ss [ 2 ] ) , ss [ 3 ] ) , ss [ 4 ] ) ) ,
                    _mm_min_epi16 ( _mm_min_epi16 ( _mm_min_epi16 ( _mm_min_epi16 (
                        ss [ 0 ] ,ss [ 1 ] ) , ss [ 2 ] ) , ss [ 3 ] ) , ss [ 4 ] ) ) ;
                _mm_store_si128 ( d - 1 , _mm_set1_epi16 ( 
                    _mm_extract_epi16 (
                        _mm_srai_epi16 (
                                _mm_adds_epi16 ( _mm_adds_epi16 (
                                    _mm_mullo_epi16 ( _mm_adds_epi16 ( wk0 , _mm_sign_epi16 ( _mm_set1_epi16 ( 2 ) , wk0 ) ) , _mm_set1_epi16 ( 42 ) ) ,
                                    _mm_srai_epi16 ( wk0 , 1 ) ) ,
                                    _mm_srai_epi16 ( wk0 , 3 ) ) ,
                                7 ) ,
                        0 ) ) ) ;

                // 
                for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , s0 ++ , s1 ++ , s2 ++ , s3 ++ , s4 ++ )
                {
                    const __m128i s [ 5 ] = {
                            _mm_load_si128 ( s0 ) ,
                            _mm_lddqu_si128 ( s1 ) , 
                            _mm_lddqu_si128 ( s2 ) ,
                            _mm_load_si128 ( s3 ) ,
                            _mm_load_si128 ( s4 ) } ;
                    const __m128i wk = _mm_subs_epi16 ( _mm_subs_epi16 (
                        _mm_adds_epi16 ( _mm_adds_epi16 ( _mm_adds_epi16 ( _mm_adds_epi16 ( 
                            s [ 0 ] ,s [ 1 ] ) , s [ 2 ] ) , s [ 3 ] ) , s [ 4 ] ) ,
                        _mm_max_epi16 ( _mm_max_epi16 ( _mm_max_epi16 ( _mm_max_epi16 ( 
                            s [ 0 ] ,s [ 1 ] ) , s [ 2 ] ) , s [ 3 ] ) , s [ 4 ] ) ) ,
                        _mm_min_epi16 ( _mm_min_epi16 ( _mm_min_epi16 ( _mm_min_epi16 ( 
                            s [ 0 ] ,s [ 1 ] ) , s [ 2 ] ) , s [ 3 ] ) , s [ 4 ] ) ) ;
                    _mm_store_si128 (
                        d ,
                        _mm_srai_epi16 (
                            _mm_adds_epi16 ( _mm_adds_epi16 (
                                _mm_mullo_epi16 ( _mm_adds_epi16 ( wk , _mm_sign_epi16 ( _mm_set1_epi16 ( 2 ) , wk ) ) , _mm_set1_epi16 ( 42 ) ) ,
                                _mm_srai_epi16 ( wk , 1 ) ) ,
                                _mm_srai_epi16 ( wk , 3 ) ) ,
                            7 ) ) ;
                }

                // E[
                _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
            }
            // [
            CopyMemory ( d - 1 , d - OneLineSpan - 1 , OneLineSpan ) ;

            // [
            d = (__m128i *) this->avg  ;
            CopyMemory ( d - 1 , d + OneLineSpan - 1 , OneLineSpan ) ;
        }
        break;
    }

    if ( this->Debug == 1 )
    {
        // ʐݒ
        const __m128i * yA = (const __m128i *) this->avg + OneLineSpan ;
        for ( int i = 0 ; i < this->Height ; i ++ , yA += 2 )
        {
            __m128i * d = (__m128i *) ( pDst + i * nSrcPitch ) ;

            // 
            for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , yA ++ )
                _mm_storeu_si128 ( d , _mm_blendv_epi8 ( _mm_lddqu_si128 ( d ) , _mm_load_si128 ( yA ) , _mm_set1_epi16 ( 0x0080 ) ) );
        }
        return src ;
    }

    {
        // gZo
        __m128i * d = (__m128i *) this->hag + OneLineSpan ;
        const __m128i * s0 = (const __m128i *) this->Y   + OneLineSpan ;
        const __m128i * s1 = (const __m128i *) this->avg + OneLineSpan ;
        for ( int i = 0 ; i < this->Height ; i ++ , d +=2 , s0 += 2 , s1 += 2 )
        {
            // [
            _mm_store_si128 ( d - 1 , _mm_set1_epi16 ( _mm_extract_epi16 ( _mm_sub_epi16 ( _mm_load_si128 ( s0 ) , _mm_load_si128 ( s1 ) ) , 0 ) ) ) ;

            // 
            for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , s0 ++ , s1 ++ )
                _mm_store_si128 ( d , _mm_sub_epi16 ( _mm_load_si128 ( s0 ) , _mm_load_si128 ( s1 ) ) ) ;

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
        }
        // [
        CopyMemory ( d - 1 , d - OneLineSpan - 1 , OneLineSpan ) ;

        // [
        d = (__m128i *) this->hag  ;
        CopyMemory ( d - 1 , d + OneLineSpan - 1 , OneLineSpan ) ;
    }

    if ( this->Debug == 2 )
    {
        // ʐݒ
        const __m128i * yA = (const __m128i *) this->hag + OneLineSpan ;
        for ( int i = 0 ; i < this->Height ; i ++ , yA += 2 )
        {
            __m128i * d = (__m128i *) ( pDst + i * nSrcPitch ) ;

            // 
            for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , yA ++ )
                _mm_storeu_si128 ( d , _mm_blendv_epi8 ( _mm_lddqu_si128 ( d ) , _mm_add_epi16 ( _mm_load_si128 ( yA ) , _mm_set1_epi16 ( 128 ) ) , _mm_set1_epi16 ( 0x0080 ) ) );
        }
        return src ;
    }

    {
        // gfBAڂŃm[}CY
        const __m128i * s0 = (const __m128i *) ( this->hag     ) + OneLineSpan ;
        const __m128i * s1 = (const __m128i *) ( this->hag - 1 ) + OneLineSpan ;
        const __m128i * s2 = (const __m128i *) ( this->hag + 1 ) + OneLineSpan ;
        const __m128i * s3 = (const __m128i *) ( this->hag     ) + OneLineSpan * 0 ;
        const __m128i * s4 = (const __m128i *) ( this->hag     ) + OneLineSpan * 2 ;
        __m128i * d = (__m128i *) this->vag + OneLineSpan ;
        for ( int i = 0 ; i < this->Height ; i ++ , d +=2 , s0 += 2 , s1 += 2 , s2 += 2 , s3 += 2 , s4 += 2 )
        {
            // [
            const __m128i ss [ 5 ] = {
                    _mm_load_si128 ( s0 ) ,
                    _mm_lddqu_si128 ( s1 ) , 
                    _mm_lddqu_si128 ( s2 ) ,
                    _mm_load_si128 ( s3 ) ,
                    _mm_load_si128 ( s4 ) } ;
            const __m128i wk0 = _mm_subs_epi16 ( _mm_subs_epi16 (
                _mm_adds_epi16 ( _mm_adds_epi16 ( _mm_adds_epi16 ( _mm_adds_epi16 ( 
                    ss [ 0 ] ,ss [ 1 ] ) , ss [ 2 ] ) , ss [ 3 ] ) , ss [ 4 ] ) ,
                _mm_max_epi16 ( _mm_max_epi16 ( _mm_max_epi16 ( _mm_max_epi16 ( 
                    ss [ 0 ] ,ss [ 1 ] ) , ss [ 2 ] ) , ss [ 3 ] ) , ss [ 4 ] ) ) ,
                _mm_min_epi16 ( _mm_min_epi16 ( _mm_min_epi16 ( _mm_min_epi16 ( 
                    ss [ 0 ] ,ss [ 1 ] ) , ss [ 2 ] ) , ss [ 3 ] ) , ss [ 4 ] ) ) ; 
            _mm_store_si128 ( d - 1 , _mm_set1_epi16 ( 
                _mm_extract_epi16 (
                    _mm_srai_epi16 (
                            _mm_adds_epi16 ( _mm_adds_epi16 (
                                _mm_mullo_epi16 ( _mm_adds_epi16 ( wk0 , _mm_sign_epi16 ( _mm_set1_epi16 ( 2 ) , wk0 ) ) , _mm_set1_epi16 ( 42 ) ) ,
                                _mm_srai_epi16 ( wk0 , 1 ) ) ,
                                _mm_srai_epi16 ( wk0 , 3 ) ) ,
                            5 ) ,
                    0 ) ) ) ;

            // 
            for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , s0 ++ , s1 ++ , s2 ++ , s3 ++ , s4 ++ )
            {
                const __m128i s [ 5 ] = {
                    _mm_load_si128 ( s0 ) ,
                    _mm_lddqu_si128 ( s1 ) , 
                    _mm_lddqu_si128 ( s2 ) ,
                    _mm_load_si128 ( s3 ) ,
                    _mm_load_si128 ( s4 ) } ;
                const __m128i wk = _mm_subs_epi16 ( _mm_subs_epi16 (
                    _mm_adds_epi16 ( _mm_adds_epi16 ( _mm_adds_epi16 ( _mm_adds_epi16 ( 
                        s [ 0 ] ,s [ 1 ] ) , s [ 2 ] ) , s [ 3 ] ) , s [ 4 ] ) ,
                    _mm_max_epi16 ( _mm_max_epi16 ( _mm_max_epi16 ( _mm_max_epi16 ( 
                        s [ 0 ] ,s [ 1 ] ) , s [ 2 ] ) , s [ 3 ] ) , s [ 4 ] ) ) ,
                    _mm_min_epi16 ( _mm_min_epi16 ( _mm_min_epi16 ( _mm_min_epi16 ( 
                        s [ 0 ] ,s [ 1 ] ) , s [ 2 ] ) , s [ 3 ] ) , s [ 4 ] ) ) ;
                _mm_store_si128 (
                    d ,
                    _mm_srai_epi16 (
                        _mm_adds_epi16 ( _mm_adds_epi16 (
                            _mm_mullo_epi16 ( _mm_adds_epi16 ( wk , _mm_sign_epi16 ( _mm_set1_epi16 ( 2 ) , wk ) ) , _mm_set1_epi16 ( 42 ) ) ,
                            _mm_srai_epi16 ( wk , 1 ) ) ,
                            _mm_srai_epi16 ( wk , 3 ) ) ,
                        5 ) ) ;
            }

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
        }
        // [
        CopyMemory ( d - 1 , d - OneLineSpan - 1 , OneLineSpan ) ;

        // [
        d = (__m128i *) this->vag  ;
        CopyMemory ( d - 1 , d + OneLineSpan - 1 , OneLineSpan ) ;
    }

    if ( this->Debug == 3 )
    {
        // ʐݒ
        const __m128i * yA = (const __m128i *) this->vag + OneLineSpan ;
        for ( int i = 0 ; i < this->Height ; i ++ , yA += 2 )
        {
            __m128i * d = (__m128i *) ( pDst + i * nSrcPitch ) ;

            // 
            for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , yA ++ )
                _mm_storeu_si128 ( d , _mm_blendv_epi8 ( _mm_lddqu_si128 ( d ) , _mm_add_epi16 ( _mm_srai_epi16 ( _mm_load_si128 ( yA ) , 2 ) , _mm_set1_epi16 ( 128 ) ) , _mm_set1_epi16 ( 0x0080 ) ) );
        }
        return src ;
    }

    {
        // 臒lȉ͍ĝĂ
        __m128i * d = (__m128i *) this->vag - 1 ;
        for ( int i = 0 ; i < ( this->Height + 2 ) * ( ( this->Width >> 3 ) + 2 ) ; i ++ , d ++ )
        {
            const __m128i wk = _mm_load_si128 ( d ) ;
            _mm_store_si128 ( d , _mm_and_si128 ( wk , _mm_cmpgt_epi16 ( _mm_abs_epi16 ( wk ) , _mm_set1_epi16 ( this->Threshold ) ) ) ) ;
        }
    }

    if ( this->Debug == 4 )
    {
        // ʐݒ
        const __m128i * yA = (const __m128i *) this->vag + OneLineSpan ;
        for ( int i = 0 ; i < this->Height ; i ++ , yA += 2 )
        {
            __m128i * d = (__m128i *) ( pDst + i * nSrcPitch ) ;

            // 
            for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , yA ++ )
                _mm_storeu_si128 ( d , _mm_blendv_epi8 ( _mm_lddqu_si128 ( d ) , _mm_add_epi16 ( _mm_srai_epi16 ( _mm_load_si128 ( yA ) , 2 ) , _mm_set1_epi16 ( 128 ) ) , _mm_set1_epi16 ( 0x0080 ) ) );
        }
        return src ;
    }

    {
        // m[}CYꂽg
        __m128i * d = (__m128i *) this->vag - 1 ;
        for ( int i = 0 ; i < ( this->Height + 2 ) * ( ( this->Width >> 3 ) + 2 ) ; i ++ , d ++ )
            _mm_store_si128 ( d , _mm_srai_epi16 ( _mm_mullo_epi16 ( _mm_load_si128 ( d ) , _mm_set1_epi16 ( this->Strength ) ) , 3 + 2 ) ) ;
    }

    if ( this->Debug == 5 )    
    {
        // ʐݒ
        const __m128i * yA = (const __m128i *) this->vag + OneLineSpan ;
        for ( int i = 0 ; i < this->Height ; i ++ , yA += 2 )
        {
            __m128i * d = (__m128i *) ( pDst + i * nSrcPitch ) ;

            // 
            for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , yA ++ )
                _mm_storeu_si128 ( d , _mm_blendv_epi8 ( _mm_lddqu_si128 ( d ) , _mm_add_epi16 ( _mm_load_si128 ( yA ) , _mm_set1_epi16 ( 128 ) ) , _mm_set1_epi16 ( 0x0080 ) ) );
        }
        return src ;
    }

    {
        // ᒆgƍg
        const __m128i * hA = (const __m128i *) this->vag + OneLineSpan ;
        const __m128i * yA = (const __m128i *) this->avg + OneLineSpan ;
        __m128i * d = (__m128i *) this->hag + OneLineSpan ;
        for ( int i = 0 ; i < this->Height ; i ++ , d += 2, hA +=2 , yA +=2 )
        {
            // [
            _mm_store_si128 ( d - 1 , _mm_set1_epi16 ( _mm_extract_epi16 ( _mm_unpacklo_epi8 ( _mm_packus_epi16 ( _mm_adds_epi16 ( _mm_load_si128 ( hA ) , _mm_load_si128 ( yA ) ) , _mm_setzero_si128 () ) , _mm_setzero_si128 () ) , 0 ) ) ) ;

            // 
            for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , hA ++ , yA ++ )
                _mm_store_si128 ( d , _mm_unpacklo_epi8 ( _mm_packus_epi16 ( _mm_adds_epi16 ( _mm_load_si128 ( hA ) , _mm_load_si128 ( yA ) ) , _mm_setzero_si128 () ) , _mm_setzero_si128 () ) ) ;

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
        }
        // [
        CopyMemory ( d - 1 , d - OneLineSpan - 1 , OneLineSpan ) ;

        // [
        d = (__m128i *) this->hag ;
        CopyMemory ( d - 1 , d + OneLineSpan - 1 , OneLineSpan ) ;
    }

    {
        // ʐݒ
        const __m128i * yA = (const __m128i *) this->hag + OneLineSpan ;
        for ( int i = 0 ; i < this->Height ; i ++ , yA += 2 )
        {
            __m128i * d = (__m128i *) ( pDst + i * nSrcPitch ) ;

            // 
            for ( int j = 0 ; j < ( this->Width >> 3 ) ; j ++ , d ++ , yA ++ )
                _mm_storeu_si128 ( d , _mm_blendv_epi8 ( _mm_lddqu_si128 ( d ) , _mm_load_si128 ( yA ) , _mm_set1_epi16 ( 0x0080 ) ) );
        }
    }
#ifdef PRINTTIME
    tcount += timeGetTime () ;
#endif // PRINTTIME
    return src ;
}
